from frontegg.baseConfig.frontegg_proxy import FronteggProxy
from flask import (Blueprint, request, make_response, Flask)
from urllib.parse import urlparse
import typing
from .secure_access import authentication_middleware as fe_auth_middleware, context_provider as fe_context_provider
from frontegg.helpers.logger import logger

class Frontegg(FronteggProxy):
    def __init__(self):
        pass

    def init_app(
            self,
            app: Flask,
            client_id: str,
            api_key: str,
            context_provider: typing.Callable = None,
            authentication_middleware = None,
            with_secure_access: bool = False,
            middleware_prefix: str = '/frontegg',
    ):
        if with_secure_access:
            authentication_middleware = authentication_middleware or fe_auth_middleware
            context_provider = context_provider or fe_context_provider

        super(Frontegg, self).__init__(client_id, api_key, context_provider, authentication_middleware,
                                       middleware_prefix)

        frontegg_blueprint = Blueprint('frontegg', __name__, url_prefix=middleware_prefix)



        @frontegg_blueprint.route('/<path:endpoint>', methods=('GET', 'POST', 'PUT', 'DELETE', 'PATCH'))
        def middleware(endpoint):

            logger.info('got new request to frontegg: method = %s, path = %s', request.method, request.path)
            o = urlparse(request.base_url)
            response = self.proxy_request(request, method=request.method, path=request.path, host=o.hostname,
                                          params=request.args, body=request.data, cookies=request.cookies,
                                          headers=dict(request.headers))

            logger.info('got response from frontegg: status_code = %s, path = %s', response.status_code, request.path)

            return make_response(response.content, response.status_code, response.headers)

        app.register_blueprint(frontegg_blueprint)

    @property
    def ignored_response_headers(self):
        return ['content-length', 'access-control-allow-credentials', 'access-control-allow-origin']




frontegg = Frontegg()


"""Frontegg client and proxy."""
from frontegg.client import FronteggRESTClient, FronteggClient, FronteggContext
from frontegg.helpers import exceptions as frontegg_exceptions, frontegg_headers
from frontegg.helpers.logger import logger as frontegg_logger


__all__ = ('FronteggRESTClient', 'FronteggClient', 'FronteggContext', 'frontegg_exceptions', 'frontegg_headers', 'frontegg_logger')



import frontegg.fastapi.frontegg as __frontegg
from fastapi import Request
from frontegg import FronteggContext
from frontegg.helpers.logger import logger

def context_provider(request: Request):
    try:
        user = __frontegg.frontegg.decode_jwt(
            request.headers.get('Authorization'))
        logger.info('decoded entity')
        return FronteggContext(user.get('sub'), user.get('tenantId'))
    except:
        return FronteggContext('user-id', 'tenant-id')


def context_provider_with_permissions(request: Request):
    try:
        user = __frontegg.frontegg.decode_jwt(
            request.headers.get('Authorization'))
        logger.info('decoded entity with permissions')
        return FronteggContext(user.get('sub'), user.get('tenantId'), user.get('permissions'))
    except:
        return FronteggContext('user-id', 'tenant-id', [])



from .context_provider import context_provider, context_provider_with_permissions
from .authentication_middleware import authentication_middleware
from .with_authentication import with_authentication

__all__ = ('context_provider', 'authentication_middleware', 'with_authentication', 'context_provider_with_permissions')


from sanic import request, response
from frontegg.baseConfig.frontegg_proxy import FronteggProxy
import typing
from .secure_access import context_provider as fe_context_provider, authentication_middleware as fe_authentication_middleware

RequestT = typing.TypeVar('RequestT')


class Frontegg(FronteggProxy):
    def __init__(self):
        pass

    async def handler(self, request, path):
        body = request.body
        host = request.headers.get('host') or request.client.host
        host = host.split(':')[0]

        response_from_proxy = self.proxy_request(request=request, method=request.method, path=path,
                                                 host=host, body=body, headers=request.headers,
                                                 params=request.args)
        return response.raw(
            response_from_proxy.content,
            headers=response_from_proxy.headers,
            status=response_from_proxy.status_code
        )

    def init_app(
            self,
            app,
            client_id: str,
            api_key: str,
            context_provider=None,
            authentication_middleware=None,
            with_secure_access: bool = None,
            middleware_prefix: str = '/frontegg',
    ):

        if with_secure_access:
            context_provider = context_provider or fe_context_provider
            authentication_middleware = authentication_middleware or fe_authentication_middleware

        super(Frontegg, self).__init__(client_id, api_key,
                                       context_provider,
                                       authentication_middleware,
                                       middleware_prefix)

        routeUrl = middleware_prefix + '/<path:[^/].*?>'
        app.add_route(self.handler, routeUrl, methods=(
            'GET', 'POST', 'PUT', 'DELETE', 'PATCH'))


frontegg = Frontegg()



from frontegg.baseConfig.frontegg_authenticator import FronteggAuthenticator
from frontegg.helpers.frontegg_urls import frontegg_urls
from urllib.parse import urljoin
import typing
from frontegg.helpers.frontegg_headers import frontegg_headers
from frontegg.helpers.exceptions import HttpException
from .identity_mixin import IdentityClientMixin
from frontegg.helpers.logger import logger


class FronteggProxy(FronteggAuthenticator, IdentityClientMixin):
    __routes_config = None

    def proxy_request(self, request, method: str, path: str, host: str, params: dict,
                      body: typing.Optional[any], cookies: typing.Optional[dict] = None,
                      headers: typing.Optional[dict] = {}):
        if not request:
            raise Exception('request is required')

        if not method:
            raise Exception('method is required')

        if not path:
            raise Exception('path is required')

        path_without_frontegg = path.replace(
            '/' + self.middleware_prefix, '', 1).replace(self.middleware_prefix, '', 1)

        logger.info('removed path prefix before sending to frontegg, new path - %s', path_without_frontegg)

        public_route = self.is_public_route(
            path_without_frontegg, params, method)

        logger.info('route is %s public route', '' if public_route else 'not')

        if self.authentication_middleware is not None and not public_route:
            try:
                logger.info('will validate user authentication')
                self.authentication_middleware(request)
            except HttpException as response:
                logger.info('failed to authorize entity')
                return response
            except Exception as e:
                logger.info('something went wrong, could not run authentication middleware')
                logger.debug('auth middleware error - %s', str(e))
                return HttpException('Something went wrong', 500)

        url = urljoin(frontegg_urls.base_url, path_without_frontegg)
        headers = self.clean_request_headers(headers, host)
        headers = self.set_context(headers, request)
        logger.info('adjusted headers before proxying request to frontegg')
        logger.debug('new headers - %s', str(headers))

        if self.should_refresh_vendor_token:
            logger.info('refresh vendor token is required, will refresh token')
            self.refresh_vendor_token()

        logger.info('will proxy request to frontegg')
        logger.debug('request data: method=%s, url=%s,cookies=%s, body=%s, params=%s, headers=%s', method, url,
                     str(cookies), str(body), str(params), str(headers))
        response = self.vendor_session_request.request(
            method,
            url,
            headers=headers,
            cookies=cookies,
            data=body,
            params=params
        )

        logger.info('got response from frontegg')

        response.headers = self.clean_response_headers(response.headers)

        return response

    def clean_request_headers(self, headers: dict, host: str) -> dict:
        new_headers = dict()
        ignored_headers = ['host', 'authorization']
        for key, value in headers.items():
            if 'access-control' not in key.lower() and key.lower() not in ignored_headers:
                new_headers[key] = value

        new_headers[frontegg_headers['vendor_host']] = host
        return new_headers

    def clean_response_headers(self, headers: dict) -> dict:
        new_headers = dict()

        ignored_headers = self.ignored_response_headers

        for key, value in headers.items():
            if key.lower() not in ignored_headers:
                new_headers[key] = value

        return new_headers

    @property
    def ignored_response_headers(self):
        return ['access-control-allow-credentials', 'access-control-allow-origin']


    def set_context(self, headers: dict, request) -> dict:
        context = self.context_callback(request)
        headers[frontegg_headers['middleware_request']] = 'True'
        if context.tenant_id is not None:
            headers[frontegg_headers['tenant_id']] = context.tenant_id
        if context.user_id is not None:
            headers[frontegg_headers['user_id']] = context.user_id
        if context.permissions is not None:
            headers[frontegg_headers['permissions']] = ','.join(context.permissions)
        return headers

    @property
    def routes_config(self):
        if self.__routes_config:
            return self.__routes_config
        response = self.vendor_session_request.get(frontegg_urls.routes_config)

        self.__routes_config = response.json()
        return self.__routes_config

    def is_public_route(self, path: str, params: dict, method: str) -> bool:
        public_routes = self.routes_config['vendorClientPublicRoutes']

        for route in public_routes:
            if path != route['url']:
                continue
            if method != route['method'].upper():
                continue
            if route.get('withQueryParams'):
                is_valid = True
                for query_param in route['withQueryParams']:
                    value = params.get(query_param['key'])
                    if not value:
                        is_valid = False
                        break
                    if query_param['value'] and value != query_param['value']:
                        is_valid = False
                        break

                if not is_valid:
                    continue
            return True
        return False



from fastapi.security.base import SecurityBase
from fastapi import Request, Depends, HTTPException
from typing import Optional, Any, Dict, List
from jwt import InvalidTokenError
import frontegg
from pydantic import BaseModel, Field
from fastapi.security.http import HTTPBearerModel
import enum


class TokenType(str, enum.Enum):
    UserToken = 'userToken'
    UserApiToken = 'userApiToken'
    TenantApiToken = 'tenantApiToken'


class User(BaseModel):
    # Fields which are general for all kind of tokens
    sub: str
    roles: List[str] = Field(default_factory=list)
    permissions: List[str] = Field(default_factory=list)
    tenant_id: str = Field(alias='tenantId')
    metadata: Dict[str, Any]
    token_type: TokenType = Field(alias='type')
    access_token: str

    # User token fields - all fields must be optional in order to support API tokens
    name: Optional[str]
    email: Optional[str]
    email_verified: Optional[bool]
    tenant_ids: Optional[List[str]] = Field(alias='tenantIds', default_factory=list)
    profile_picture_url: Optional[str] = Field(alias='profilePictureUrl')

    # API Token fields - all fields must be optional in order to support user tokens
    created_by_user_id: Optional[str] = Field(alias='createdByUserId')

    def has_permissions(self, permissions: List[str]) -> bool:
        return bool(permissions) and all(p in self.permissions for p in permissions)

    @property
    def id(self) -> Optional[str]:
        """
        When using tenant API Token, there is no user ID.
        When using user API Token, the user ID specified in the created_by_user_id field.
        Otherwise, the user ID is specified in the sub field.
        """
        if self.token_type == TokenType.TenantApiToken:
            return None

        return self.created_by_user_id or self.sub


class FronteggHTTPAuthentication(SecurityBase):
    def __init__(self,
                 bearerFormat: Optional[str] = None,  # noqa
                 scheme_name: str = None,
                 auto_error: bool = True):
        self.model = HTTPBearerModel(bearerFormat=bearerFormat)
        self.scheme_name = scheme_name or self.__class__.__name__
        self.auto_error = auto_error

    async def __call__(self, request: Request) -> Optional[User]:
        authorization: str = request.headers.get("Authorization")
        try:
            decoded = frontegg.fastapi.frontegg.decode_jwt(authorization)
            return User(**decoded, access_token=authorization.replace('Bearer ', ''))
        except InvalidTokenError:
            if self.auto_error:
                raise HTTPException(
                    status_code=401, detail="Unauthenticated"
                )
            else:
                return None


def FronteggSecurity(permissions: List[str] = None, auto_error: bool = True):  # noqa
    """
    This factory function will create authentication dependency for FastAPI,
    and will ensure the user has the right permissions if specified.
    """

    def check_perm(user: User = Depends(FronteggHTTPAuthentication(auto_error=auto_error))):
        if permissions and not user.has_permissions(permissions=permissions):
            raise HTTPException(status_code=403, detail='You do not have permission to perform this action.')
        return user

    return check_perm



from ..helpers.logger import logger


class FronteggConfig():
    client_id: str = None
    api_key: str = None

    def __init__(self, client_id: str, api_key, context_provider=None,
                 authentication_middleware=None, middleware_prefix: str = None):
        if client_id is None:
            raise Exception('client_id is required')
        if api_key is None:
            raise Exception('api_key is required')
        if context_provider is None:
            raise Exception('context_provider is required')
        if authentication_middleware is None:
            logger.warning('authentication middleware was not provided. In order to protect frontegg routes, it is recommended to provide authentication_middleware which validates the authentication of the user.')

        self.client_id = client_id
        self.api_key = api_key
        self.context_callback = context_provider
        self.authentication_middleware = authentication_middleware

        # why call "frontegg/" as well? to enable overriding the fix_middleware_prefix func with suitable func for the web framework
        self.middleware_prefix = self.fix_middleware_prefix_format(
            middleware_prefix or 'frontegg/')

    def fix_middleware_prefix_format(self, prefix: str):
        if prefix.startswith('/'):
            prefix = prefix[1:]
        if not prefix.endswith('/'):
            prefix += '/'

        return prefix



import frontegg.sanic as __frontegg
from sanic import request
from frontegg import FronteggContext



def context_provider(request: request):
    try:
        user = __frontegg.frontegg.decode_jwt(request.headers.get('Authorization'))
        return FronteggContext(user.get('sub'), user.get('tenantId'))
    except:
        return FronteggContext('user-id', 'tenant-id')


def context_provider_with_permissions(request: request):
    try:
        user = __frontegg.frontegg.decode_jwt(request.headers.get('Authorization'))
        return FronteggContext(user.get('sub'), user.get('tenantId'), user.get('permissions'))
    except:
        return FronteggContext('user-id', 'tenant-id', [])


import typing
from functools import wraps
import frontegg.flask as __frontegg
from flask import request, abort
from frontegg.helpers.logger import logger
from flask import g


def with_authentication(
        permission_keys: typing.Optional[list] = None,
        role_keys: typing.Optional[list] = None
):
    def decorator(f):
        @wraps(f)
        def decorated_function(*args, **kwargs):
            # Initially
            valid_permissions = True
            valid_roles = True
            try:
                decoded = __frontegg.frontegg.decode_jwt(request.headers.get('Authorization'))
                g.user = decoded
                # Validate roles
                if role_keys is not None:
                    logger.info('will check if entity has one of required roles')
                    valid_roles = any(
                        role in decoded['roles'] for role in role_keys)

                if permission_keys is not None:
                    logger.info('will check if entity has one of required permissions')
                    valid_permissions = any(
                        permission in decoded['permissions'] for permission in permission_keys)

            except Exception as e:
                logger.debug('something went wrong while validating roles and permissions, ' + str(e))
                abort(401)

            if not valid_permissions or not valid_roles:
                logger.info('entity does not have required role and permissions')
                abort(403)
                return

            logger.info('entity passed authentication middleware')

            return f(*args, **kwargs)

        return decorated_function

    return decorator



from .security import FronteggHTTPAuthentication, FronteggSecurity, User
from .context_provider import context_provider, context_provider_with_permissions
from .authentication_middleware import authentication_middleware

__all__ = ('FronteggHTTPAuthentication', 'FronteggSecurity', 'context_provider', 'authentication_middleware', 'User',
           'context_provider_with_permissions')



frontegg_headers = {
    'vendor_host': 'frontegg-vendor-host',
    'tenant_id': 'frontegg-tenant-id',
    'user_id': 'frontegg-user-id',
    'permissions': 'frontegg-user-permissions',
    'middleware_request': 'frontegg-middleware-request',
}


from abc import ABCMeta, abstractmethod
from frontegg.helpers.frontegg_urls import frontegg_urls
import typing
import jwt
import requests
from frontegg.helpers.logger import logger


class IdentityClientMixin(metaclass=ABCMeta):
    __publicKey = None

    @property
    @abstractmethod
    def vendor_session_request(self) -> requests.Session:
        pass

    @property
    @abstractmethod
    def should_refresh_vendor_token(self) -> bool:
        pass

    @abstractmethod
    def refresh_vendor_token(self) -> None:
        pass

    def get_public_key(self) -> str:
        if self.__publicKey:
            return self.__publicKey

        logger.info('could not find public key locally, will fetch public key')
        reties = 0
        while reties < 10:
            try:
                self.__publicKey = self.fetch_public_key()
                return self.__publicKey
            except Exception as e:
                reties = reties + 1
                logger.error(
                    'could not get public key from frontegg, retry number - ' + str(reties) + ', ' + str(e))

        logger.error('failed to get public key in all retries')

    def fetch_public_key(self) -> str:

        if self.should_refresh_vendor_token:
            self.refresh_vendor_token()

        response = self.vendor_session_request.get(
            frontegg_urls.identity_service['vendor_config'])
        response.raise_for_status()
        data = response.json()
        return data.get('publicKey')

    def decode_jwt(self, authorization_header, verify: typing.Optional[bool] = True):
        if not authorization_header:
            raise Exception('Authorization headers is missing')
        logger.debug('found authorization header: ' +
                     str(authorization_header))
        jwt_token = authorization_header.replace('Bearer ', '')
        if verify:
            public_key = self.get_public_key()
            logger.debug('got public key' + str(public_key))
            decoded = jwt.decode(jwt_token, public_key, algorithms='RS256')
        else:
            decoded = jwt.decode(jwt_token, algorithms='RS256', verify=False)

        logger.info('jwt was decoded successfully')
        logger.debug('JWT value - ' + str(decoded))
        return decoded



from sanic import request
from frontegg.helpers.exceptions import UnauthenticatedException
import frontegg.sanic as __frontegg


def authentication_middleware(request: request):
    try:
        __frontegg.frontegg.decode_jwt(
            request.headers.get('Authorization'))
        return None
    except:
        raise UnauthenticatedException()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['frontegg',
 'frontegg.baseConfig',
 'frontegg.fastapi',
 'frontegg.fastapi.secure_access',
 'frontegg.flask',
 'frontegg.flask.secure_access',
 'frontegg.helpers',
 'frontegg.sanic',
 'frontegg.sanic.secure_access']

package_data = \
{'': ['*']}

install_requires = \
['arrow>=0.15.5,<0.16.0',
 'cryptography>=3.1,<4.0',
 'pyjwt>=1.7.1,<2.0.0',
 'requests>=2.22.0,<3.0.0']

extras_require = \
{u'fastapi': ['fastapi'], u'flask': ['flask>=1.0,<2.0']}

setup_kwargs = {
    'name': 'frontegg',
    'version': '1.0.6',
    'description': 'Frontegg is a web platform where SaaS companies can set up their fully managed, scalable and brand aware - SaaS features and integrate them into their SaaS portals in up to 5 lines of code.',
    'long_description': '.. image:: https://fronteggstuff.blob.core.windows.net/frongegg-logos/logo-transparent.png\n   :alt: Frontegg\n\nFrontegg is a web platform where SaaS companies can set up their fully managed, scalable and brand aware - SaaS features and integrate them into their SaaS portals in up to 5 lines of code.\n\nFor more information and usage you can visit the `github repo <https://github.com/frontegg/python-sdk>`_.',
    'author': 'Frontegg LTD',
    'author_email': 'hello@frontegg.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://frontegg.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from .context_provider import context_provider, context_provider_with_permissions
from .authentication_middleware import authentication_middleware
from .with_authentication import with_authentication

__all__ = ('context_provider',
           'authentication_middleware',
           'with_authentication',
           'context_provider_with_permissions')



from flask import Request
from frontegg.helpers.exceptions import UnauthenticatedException
from frontegg.helpers.logger import logger
import frontegg.flask as __frontegg


def authentication_middleware(request: Request):
    try:
        __frontegg.frontegg.decode_jwt(request.headers.get('Authorization'))
        logger.info('JWT token verified')
        return None
    except Exception as e:
        logger.debug('could not verify JWT token, ' + str(e), )
    raise UnauthenticatedException()



from fastapi import FastAPI, Request, Response
from frontegg.baseConfig.frontegg_proxy import FronteggProxy
import typing
import frontegg.fastapi.secure_access as secure_access
from fastapi.concurrency import run_in_threadpool


class Frontegg(FronteggProxy):
    def __init__(self):
        pass

    def init_app(
            self,
            app: FastAPI,
            client_id: str,
            api_key,
            context_provider: typing.Callable = None,
            authentication_middleware=None,
            with_secure_access: bool = False,
            middleware_prefix: str = None,
    ):

        if with_secure_access:
            context_provider = context_provider or secure_access.context_provider
            authentication_middleware = authentication_middleware or secure_access.authentication_middleware

        super(Frontegg, self).__init__(client_id, api_key, context_provider, authentication_middleware, middleware_prefix)

        @app.api_route(
            path="/" + self.middleware_prefix + "{application_path:path}",
            methods=["GET", "POST", "PUT", "PATCH", "DELETE", "OPTIONS"],
            include_in_schema=False,
        )
        async def proxy_frontegg_requests(
                application_path: str,
                request: Request,
        ) -> Response:
            body = await request.body()
            host = request.headers.get('host') or request.client.host
            host = host.split(':')[0]
            response = await run_in_threadpool(
                self.proxy_request,
                request=request, method=request.method, path=application_path,
                host=host, body=body, headers=request.headers,
                params=request.query_params
            )
            return Response(content=response.content, status_code=response.status_code, headers=response.headers)


frontegg = Frontegg()
